unit SEMain;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ComCtrls, ShlObj;

type
  TFrmMain = class(TForm)
    grpFile   : TGroupBox;
    lblPAK    : TLabel;
    edtPak    : TEdit;
    btnPak    : TButton;
    lblExt    : TLabel;
    edtExt    : TEdit;
    btnExt    : TButton;
    grpCom    : TGroupBox;
    btnExtract: TButton;
    stsInfo   : TStatusBar;
    prgInfo   : TProgressBar;
    grpLog    : TGroupBox;
    memLog    : TMemo;
    dlgOpen   : TOpenDialog;
    procedure btnPakClick(Sender: TObject);
    procedure btnExtClick(Sender: TObject);
    procedure btnExtractClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
  private
    { Private declarations }
  public
    Stop    : Boolean;
    pakType : String;
    pakCount: Integer;
    lenCount: Integer;
    //-----------------------------
    procedure Msg(Text: String);
    procedure ShowInfo(k: Integer; s: String);
    procedure WriteLog(sPath,s: String);
    function  IsValidFile(hFile: TFileStream): Boolean;
    function  GetNameById(i: Integer): String;
    function  GetExtByType(typ: Integer): String;
    function  SaveFile(hFile: TFileStream;  sPath: String; i,typ,ofs,len: Integer): String;    
    procedure Extract(hFile: TFileStream; sPath: String);
    procedure StartExtract(sFile, sPath: String);
  end;

var
  FrmMain: TFrmMain;

implementation

{$R *.dfm}

function GetFolder(Wnd: HWND; Title: String; bifType:Integer): String;
var
  lpItemID: PItemIDList;
  BrowseInfo: TBrowseInfo;
  Buf: array [0..MAX_PATH] of Char;
begin
  Result:='';
  FillChar(BrowseInfo, SizeOf(TBrowseInfo), #0);
  BrowseInfo.hwndOwner:=Wnd;
  BrowseInfo.pszDisplayName:=@Buf;
  BrowseInfo.lpszTitle:=PChar(Title);
  BrowseInfo.ulFlags:=bifType;
  lpItemId:=SHBrowseForFolder(BrowseInfo);
  if lpItemId<>nil then
  begin
    if SHGetPathFromIDList(lpItemId, Buf) then Result:=Buf;
    GlobalFreePtr(lpItemID);
  end;
end;

procedure TFrmMain.Msg(Text: String);
begin
  Application.MessageBox(PAnsiChar(Text), ''); 
end;

procedure TFrmMain.ShowInfo(k: Integer; s: String);
begin
  prgInfo.Position:=k;
  stsInfo.SimpleText:=s;
  memLog.Lines.Add(s);
  Application.ProcessMessages; 
end;

procedure TFrmMain.WriteLog(sPath,s: String);
const
  log = 'Extract.log';
var
  F: TextFile;
  sx: String;
begin
  sx:=sPath+'\'+log;
  AssignFile(F, sx);
  if FileExists(sx) then Append(F) else Rewrite(F);
  Writeln(F,s);
  CloseFile(F);
end;

function TFrmMain.IsValidFile(hFile: TFileStream): Boolean;
const
  max = 3;
  mas: array [1..max] of String = ('SND','TEX','MDL');
var
  s: String;
  c: Char;
  isValid: Boolean;
  i: Integer;
begin
  s:=''; pakType:=''; pakCount:=0; lenCount:=0;
  hFile.Seek(0,soFromBeginning);
  hFile.Read(c,1); s:=s+c;
  hFile.Read(c,1); s:=s+c;
  hFile.Read(c,1); s:=s+c;
  i:=1; isValid:=false;
  while (i<=max) and not isValid do
  begin
    if mas[i]=s then
    begin
      pakType:=s; isValid:=true;
    end;
    i:=i+1;
  end;
  if isValid then
  begin
    hFile.Seek(4,soFromBeginning);
    hFile.Read(pakCount,4);
    lenCount:=Length(IntToStr(pakCount));
    prgInfo.Position:=0; prgInfo.Max:=pakCount;
  end;
  Result:=isValid;
end;

function TFrmMain.GetNameById(i: Integer): String;
var
  s: String;
begin
  s:=IntToStr(i); while Length(s)<lenCount do s:='0'+s;
  Result:=s;
end;

function TFrmMain.GetExtByType(typ: Integer): String;
begin
  Result:='.'+IntToHex(typ,2);
  case typ of
    $20: Result:='.wav';
    $21: Result:='.mp3';
  end;
end;

function TFrmMain.SaveFile(hFile: TFileStream; sPath: String; i,typ,ofs,len: Integer): String;
var
  s,sr: String;
  buf: PChar;
  wFile: TFileStream;
begin
  GetMem(buf,len);
  try
    hFile.Seek(ofs,soFromBeginning);
    hFile.Read(buf^,len);
    sr:=GetNameById(i) + GetExtByType(typ);
    s:=sPath+'\'+sr;
    ShowInfo(i,s);
    wFile:=TFileStream.Create(s,fmCreate);
    try
      wFile.Write(buf^,len);
    finally
      wFile.Free;
    end;
  finally
    FreeMem(buf);
  end;
  Result:=sr;
end;

procedure TFrmMain.Extract(hFile: TFileStream; sPath: String);
var
  i: Integer;
  typ,ofs,len: Integer;
  sek: Int64;
  s: String;
begin
  sek:=$100; i:=1;
  while (i<=pakCount) and not Stop do
  begin
    hFile.Seek(sek,soFromBeginning);
    hFile.Read(typ,4);
    hFile.Read(ofs,4);
    hFile.Read(len,4);
    sek:=hFile.Position;
    if (typ<>0) and (ofs<>0) and (len<>0) then
    begin
      s:=SaveFile(hFile,sPath,i,typ,ofs,len);
      s:=s + ' [' + IntToHex(sek-$C,8) + ']: '+IntToHex(ofs,8) + ', ' + IntToHex(len,8);
      WriteLog(sPath,s);
    end;
    i:=i+1;
  end;
end;

procedure TFrmMain.StartExtract(sFile, sPath: String);
var
  hFile: TFileStream;
begin
  if FileExists(sFile) then
  begin
    hFile:=TFileStream.Create(sFile,fmOpenRead);
    try
      if IsValidFile(hFile) then
      begin
        ForceDirectories(sPath);
        WriteLog(sPath,'=== Start ===');
        Extract(hFile,sPath);
        WriteLog(sPath,'=== Finish ===');        
      end
      else Msg('  PAK-   :(');
    finally
      hFile.Free;
    end;
  end
  else Msg('   [' + sFile + ']');
end;

procedure TFrmMain.btnPakClick(Sender: TObject);
begin
  if dlgOpen.Execute then edtPak.Text:=dlgOpen.FileName;
end;

procedure TFrmMain.btnExtClick(Sender: TObject);
var
  s: String;
begin
  s:=GetFolder(Application.Handle, 'Path', 0);
  if Length(s)>0 then edtExt.Text:=s;
end;

procedure TFrmMain.btnExtractClick(Sender: TObject);
var
  ret: Integer;
begin
  if Stop then
  begin
    Stop:=false;
    memLog.Clear;
    btnExtract.Caption:='';
    //------------------------------------
    StartExtract(edtPak.Text, edtExt.Text);
    //------------------------------------
    btnExtract.Caption:='';
    prgInfo.Position:=0; stsInfo.SimpleText:='';
    Stop:=true;
  end else
  begin
    ret:=Application.MessageBox(' ,    ?','',MB_YESNO);
    if ret=ID_YES then Stop:=true; 
  end;
end;

procedure TFrmMain.FormCreate(Sender: TObject);
begin
  Stop:=true;
end;

end.
